import contextlib
import itertools
import threading
import time
import weakref

from dogpile.cache.backends import memcached as memcached_backend
import memcache


class ConnectionPool(object):
    def __init__(self, maxsize, unused_timeout):
        self._maxsize = maxsize
        self._unused_timeout = unused_timeout
        self._acquired = 0
        self._free_pool = []
        self._lock = threading.Lock()
        no_free_cond = self._no_free = threading.Condition(self._lock)
        stop_event = threading.Event()

        def stop(ref):
            stop_event.set()
            # Notify cleaner thread if its waiting on empty pool
            with no_free_cond:
                no_free_cond.notify_all()
        self._ref = weakref.ref(self, stop)

        cleaner_thread = threading.Thread(
            target=self._cleaner,
            args=(self._free_pool, stop_event, self._no_free),
        )
        cleaner_thread.daemon = True
        cleaner_thread.start()

    def _create_connection(self):
        raise NotImplementedError

    # Need this to be available without reference to self to call it from
    # cleaner thread
    @classmethod
    def _destroy_connection(cls, conn):
        raise NotImplementedError

    def acquire(self):
        with self._lock:
            # Wait for either free connection or free space for a new one
            while not self._free_pool and self._acquired >= self._maxsize:
                self._no_free.wait()
            try:
                # Hope there's a free one
                _, conn = self._free_pool.pop()
            except IndexError:
                # Create a new one
                conn = self._create_connection()
            self._acquired += 1
            return conn

    def release(self, conn):
        with self._lock:
            # Return connection to pool
            self._free_pool.append((time.time() + self._unused_timeout, conn))
            self._acquired -= 1
            # Notify waiting getters and cleaner
            self._no_free.notify_all()

    @contextlib.contextmanager
    def get(self):
        conn = self.acquire()
        try:
            yield conn
        finally:
            self.release(conn)

    @classmethod
    def _cleaner(cls, free_pool, stop_event, no_free_cond):
        # We'll exit when stop_event is fired and the pool is empty
        while True:
            with no_free_cond:
                # If pool is empty, wait for new connection there
                while not free_pool and not stop_event.is_set():
                    no_free_cond.wait()
                # We have at least until timeout of the bottom conn to sleep
                if not stop_event.is_set():
                    sleep_until = free_pool[0][0]
                else:
                    if not free_pool:
                        break  # We're being stopped and the pool is empty
                    else:
                        sleep_until = 0  # Don't sleep, just kill'em all
            # Release lock during sleep to let pool work
            now = time.time()
            if now < sleep_until:
                stop_event.wait(sleep_until - now)
            with no_free_cond:
                to_close = None
                # If we're stopping we need to close all connections
                if not stop_event.is_set():
                    now = time.time()
                    # Find first connection that hasn't timed out yet
                    for i, (timeout, conn) in enumerate(free_pool):
                        if timeout >= now:
                            to_close = free_pool[:i]
                            free_pool[:i] = []
                            break
                if to_close is None:
                    # We are closing or didn't find not timed out connection,
                    # so we need to close all connections (if there are any)
                    to_close = free_pool[:]
                    free_pool[:] = []
            for _, conn in to_close:
                cls._destroy_connection(conn)
            del to_close


class DumbPool(ConnectionPool):
    def __init__(self, maxsize, unused_timeout):
        super(DumbPool, self).__init__(maxsize, unused_timeout)
        self._count = itertools.count()

    def _create_connection(self):
        conn = next(self._count)
        print "Created", conn
        return conn

    @classmethod
    def _destroy_connection(cls, conn):
        print "Destroyed", conn

# This 'class' is taken from http://stackoverflow.com/a/22520633/238308
# Don't inherit client from threading.local so that we can reuse clients in
# different threads
MemcacheClient = type('MemcacheClient', (object,),
                      dict(memcache.Client.__dict__))


class MemcacheClientPool(ConnectionPool):
    def __init__(self, url, **kwargs):
        super(MemcacheClientPool, self).__init__(**kwargs)
        self.url = url

    def _create_connection(self):
        return MemcacheClient(self.url)

    @classmethod
    def _destroy_connection(cls, conn):
        conn.disconnect_all()


# Helper to ease backend refactoring
class ClientProxy(object):
    def __init__(self, client_pool):
        # Get connection from pool
        client = self.client = client_pool.acquire()
        # Release it once no reference to proxy left
        self._ref = weakref.ref(self, lambda ref: client_pool.release(client))

    def __getattr__(self, name):
        return getattr(self.client, name)


class PooledMemcachedBackend(memcached_backend.MemcachedBackend):
    # Composed from GenericMemcachedBackend's and MemcacheArgs's __init__
    def __init__(self, arguments):
        super(PooledMemcachedBackend, self).__init__(arguments)
        self.client_pool = MemcacheClientPool(
            self.url,
            maxsize=arguments.get('pool_maxsize', 10),
            unused_timeout=arguments.get('pool_unused_timeout', 60),
        )

    # Since all methods in backend just call one of methods of client, this
    # lets us avoid need to hack it too much
    @property
    def client(self):
        return ClientProxy(self.client_pool)
